// The Mini Wall 1.0 by MagicManu
// 06/2017
// Video : https://www.youtube.com/watch?v=EzuBt4-UgCc
// Blog : http://blog.magicmanu.com


// MD_MAX72xx and MD_KeySwitch library https://github.com/MajicDesigns
#include <MD_MAX72xx.h>
#include <MD_KeySwitch.h>

// RGBLED library https://github.com/VaSe7u/RGBLED
#include <RGBLED.hpp>

//#include <SPI.h>

// 4 MAX7219 displays
#define MAX_DEVICES 4
#define CLK_PIN   12
#define DATA_PIN  11
#define CS_PIN    10
#define CHAR_SPACING  1   // Pixels between characters
#define BUF_SIZE      75  // character buffer size
uint32_t prevTimeAnim = 0;  // Used for remembering the millis() value in animations
#define SCROLL_DELAY    30
static  boolean bRestart = true;
MD_MAX72XX mx = MD_MAX72XX(DATA_PIN, CLK_PIN, CS_PIN, MAX_DEVICES);


// Switchs
MD_KeySwitch BtGO(7, LOW);  // GO switch
MD_KeySwitch BtMO(6, LOW);  // Motor switch

// Buttons
MD_KeySwitch BtPL(8, LOW);  // +  button
MD_KeySwitch BtOK(9, LOW);  // OK button


// RGB Led strip
const byte redStrip = A3;
const byte greenStrip = A4;
const byte blueStrip = A5;
RGBLED rgbStrip(redStrip, greenStrip, blueStrip, CommonElectrode::cathode);


// 74HC165 shift registers x2 (16 limit switch)
int LATCH = 4;  // pin 1 (Load)
int CLOCK = 3;  // pin 2 (Clock)
int DATA = 2;   // pin 9 (Q7)
int Switch[16]; // Buttons[1] to Buttons[8] (LOW or HIGH)


// Buzzer
int outBuzzer = 5;//A0;


// Motor
int outMotor = A2;


// Cycle, game
char Tampon[40];    // Temporary value for display
int STEP = 0;       // Current step
int NbPlayer = 1;   // Number of player
int CurPlayer = 1;  // Actual player 1-8
int NbTurn = 7;     // Number of round in game
int CurTurn = 1;    // Actual round game 1-10
int Scores[9];      // Total score of each player
int FinalScore = 1; //
int IdWinner = 1;   // Num of winner
int ValWinner = 0;  // Score of winner
int NbBallFall = 0; // Number of ball fall in a round
int NbBallGiven = 0;// Number of balls given
int EndLoop = 1;    // End for flashing colors
int PrevScore = 0;  // Score before news points
int StepDispl = 1;  // Increment of the loop for display score

long TimeRef = 0;   // Time at the start of change motor state
int Cpt = 0;        // Number of millisecond of the motor start or stop
boolean StateMotor = LOW;
int MaxMSMotor = 0;

long TimeRefBallFall = 0;   // Time at the start of go button is pressed
int CptBallFall = 0;        // Number of millisecond of the ball fell

int NbBalls[] = {1, 1, 1, 2, 2, 3}; // Random number of ball but more chance to have 1 or 2 balls
int Color[] = {1, 1, 1, 2, 2}; // 1=GREEN, 2=RED Random color but more chance to have greens balls

int NbBallsSelect = 0;  // Number of balls for the next player
int ColorSelect = 0;    // Color for the next player


// Points of each limit switch
int Points[] = {
  10,
  100,
  1,
  500,
  10,
  1000,
  1,
  0,
  1,
  500,
  100,
  10,
  50,
  400,
  1,
  200
};

// FR Version
char* TxtArray[] =
{
  "The Mini Wall",        // 0
  "Nombre de joueurs ?",  // 1
  "%d Jou.",              // 2
  "Jou. %d :",            // 3
  "Termine !",            // 4
  "Joueur %d=%d  ",       // 5
  "SCORES : ",            // 6
  "Manche %d",            // 7
  "BRAVO joueur %d !!!"   // 8
};

/*
// EN / US Version
char* TxtArray[] =
{
  "The Mini Wall",        // 0
  "Number of player?",    // 1
  "%d play.",             // 2
  "Play. %d :",           // 3
  "Finished!",            // 4
  "Player %d=%d  ",       // 5
  "SCORES : ",            // 6
  "Step %d",              // 7
  "Congratulation player %d !!!"   // 8
};
*/


void printText(char *pMsg){
// Print the text string to the LED matrix modules specified.
// Message area is padded with blank columns after printing.

  uint8_t   modStart = 0;
  uint8_t   modEnd = 3;
  uint8_t   state = 0;
  uint8_t   curLen;
  uint16_t  showLen;
  uint8_t   cBuf[8];
  int16_t   col = ((modEnd + 1) * COL_SIZE) - 1;

  mx.control(modStart, modEnd, MD_MAX72XX::UPDATE, MD_MAX72XX::OFF);

  do     // finite state machine to print the characters in the space available
  {
    switch(state)
    {
      case 0: // Load the next character from the font table
        // if we reached end of message, reset the message pointer
        if (*pMsg == '\0')
        {
          showLen = col - (modEnd * COL_SIZE);  // padding characters
          state = 2;
          break;
        }

        // retrieve the next character form the font file
        showLen = mx.getChar(*pMsg++, sizeof(cBuf)/sizeof(cBuf[0]), cBuf);
        curLen = 0;
        state++;
        // !! deliberately fall through to next state to start displaying

      case 1: // display the next part of the character
        mx.setColumn(col--, cBuf[curLen++]);

        // done with font character, now display the space between chars
        if (curLen == showLen)
        {
          showLen = CHAR_SPACING;
          state = 2;
        }
        break;

      case 2: // initialize state for displaying empty columns
        curLen = 0;
        state++;
        // fall through

      case 3: // display inter-character spacing or end of message padding (blank columns)
        mx.setColumn(col--, 0);
        curLen++;
        if (curLen == showLen)
          state = 0;
        break;

      default:
        col = -1;   // this definitely ends the do loop
    }
  } while (col >= (modStart * COL_SIZE));

  mx.control(modStart, modEnd, MD_MAX72XX::UPDATE, MD_MAX72XX::ON);
}



bool scrollText(bool bInit, char *pmsg)
// Callback function for data that is required for scrolling into the display
{
  static char    curMessage[BUF_SIZE];
  static char   *p = curMessage;
  static uint8_t  state = 0;
  static uint8_t  curLen, showLen;
  static uint8_t  cBuf[8];
  uint8_t         colData;

  // are we initializing?
  if (bInit)
  {
    resetMatrix();
    strcpy(curMessage, pmsg);
    state = 0;
    p = curMessage;
    bInit = false;
  }

  // Is it time to scroll the text?
  if (millis()-prevTimeAnim < SCROLL_DELAY)
    return(bInit);

  // scroll the display
  mx.transform(MD_MAX72XX::TSL);  // scroll along
  prevTimeAnim = millis();      // starting point for next time

  // now run the finite state machine to control what we do
  switch (state)
  {
    case 0: // Load the next character from the font table
      showLen = mx.getChar(*p++, sizeof(cBuf)/sizeof(cBuf[0]), cBuf);
      curLen = 0;
      state = 1;

    // !! deliberately fall through to next state to start displaying

    case 1: // display the next part of the character
      colData = cBuf[curLen++];
      mx.setColumn(0, colData);
      if (curLen == showLen)
      {
        showLen = ((*p != '\0') ? CHAR_SPACING : mx.getColumnCount()-1);
        curLen = 0;
        state = 2;
      }
      break;

    case 2: // display inter-character spacing (blank column) or scroll off the display
      mx.setColumn(0, 0);
      if (++curLen == showLen)
      {
        state = 0;
        bInit = (*p == '\0');
      }
      break;

    default:
      state = 0;
  }

  return(bInit);
}


void resetMatrix(void)
{
  mx.control(MD_MAX72XX::INTENSITY, MAX_INTENSITY/2);
  mx.control(MD_MAX72XX::UPDATE, MD_MAX72XX::ON);
  mx.clear();
  prevTimeAnim = 0;
}


void setup()
{ 
  // Init Buttons
  BtGO.begin();
  BtMO.begin();
  BtPL.begin();
  BtOK.begin();
  BtGO.enableDoublePress(false);
  BtMO.enableDoublePress(false);
  BtPL.enableDoublePress(false);
  //BtOK.enableDoublePress(false);

  // LED strip blue on start
  rgbStrip.on();
  rgbStrip.brightness(PWM_MAX);
  rgbStrip(0, 0, 255);
  rgbStrip.show();
  
  // 74HC165
  pinMode(LATCH, OUTPUT);
  pinMode(CLOCK, OUTPUT);
  pinMode(DATA, INPUT);
  digitalWrite(CLOCK, LOW);
  digitalWrite(LATCH, LOW);  
  
  // Init Display
  mx.begin();
  prevTimeAnim = millis();

  // Buzzer
  pinMode(outBuzzer, OUTPUT); 
   
  // Motor
  pinMode(outMotor, OUTPUT);  

  // For generate random values
  randomSeed(analogRead(0));

  // Start Bip
  tone(outBuzzer, 3000, 600);
}


void loop()
{ 
  
  // The complete steps of game
  switch (STEP)
  {
    case 0:
      // Init for new game
      CurPlayer = 1;
      CurTurn = 1; 
      FinalScore = 1;
      ValWinner = 0;
      NbBallFall = 0;
      NbBallGiven = 0;
      memset(Scores, 0, sizeof(Scores));  // reset scores
      rgbStrip(0, 0, 255);  // Blue
      
      // Boot : display name game
      if (bRestart = scrollText(bRestart, TxtArray[0]))
        STEP = 10;
      break;

      
    case 10:
      // Ask number of player
      if (bRestart = scrollText(bRestart, TxtArray[1]))
        STEP = 20;
      break;

      
    case 20:
      // Display the number of player
      sprintf(Tampon, TxtArray[2], (int) NbPlayer);  
      printText(Tampon);

      // Increment players with +
      if (BtPL.read() == MD_KeySwitch::KS_PRESS){
        NbPlayer++;
        if (NbPlayer > 8)
          NbPlayer = 1;    
      }

      // Valid number of players with OK 
      if (BtOK.read() == MD_KeySwitch::KS_PRESS)
        STEP = 30;
      break;


    case 30:
      // Show current round of game
      sprintf(Tampon, TxtArray[7], (int) CurTurn);   
      if (bRestart = scrollText(bRestart, Tampon)){

        NbBallGiven = 0;

        // How many balls ?
        if (CurTurn < 3)
          NbBallsSelect = 1;  // Always ONE ball for the first two rounds
        else
          NbBallsSelect = NbBalls[random(0, 6)]; // Give 1 to 3 balls         


        TimeRef = millis();
        STEP = 40;
      }
      break;   


    case 40:
      // Distributes the balls
      rgbStrip(0, 0, 0);  // Black
      rgbStrip.show();    

      // Show number of balls
      sprintf(Tampon, "X %d", (int) NbBallsSelect);  
      printText(Tampon);

   
      // Turn motor : cutting the signal to reduce the speed
      Cpt = millis() - TimeRef;
      MaxMSMotor = StateMotor ? 10 : 15;
      if (Cpt > MaxMSMotor){
        StateMotor = !StateMotor;
        TimeRef = millis();
      }
      digitalWrite(outMotor, StateMotor);


      // Limit switch of motor for count balls
      if (BtMO.read() == MD_KeySwitch::KS_PRESS)
        NbBallGiven++;
              
      if (NbBallGiven >= NbBallsSelect){
        digitalWrite(outMotor, LOW);
        STEP = 50;      
      }
      break; 

      
    case 50:
      // Choise color
      if (CurTurn < 3)
        ColorSelect = 1; // Always green for the first two rounds
      else
        ColorSelect = Color[random(0, 5)];  // 1=GREEN OR 2=RED ?

      EndLoop = ColorSelect == 1 ? 11 : 10;
    
      // Flashes colors (suspens !)
      for (int i = 0; i < EndLoop; i++){
        if (i%2 == 0)
          rgbStrip(0, 255, 0);  // Green   
        else
          rgbStrip(255, 0, 0);  // Red
        rgbStrip.show();    
        delay(200);       
      }
        
      STEP = 60;
      break;   
          

    case 60:   
      // Ask player X to play
      sprintf(Tampon, TxtArray[3], (int) CurPlayer);            
      printText(Tampon);

      NbBallFall = 0;
      
      // Valid player when GO button is pressed
      if (BtGO.read() == MD_KeySwitch::KS_PRESS){
        TimeRefBallFall = millis();
        STEP = 70; 
      }        
      break;

      
    case 70:   
      // Read limit switchs
      digitalWrite(LATCH, HIGH);
      for (int i = 0; i <= 15; i++){
        Switch[15-i] = digitalRead(DATA);
        digitalWrite(CLOCK, HIGH);
        digitalWrite(CLOCK, LOW);   
      } 
      digitalWrite(LATCH, LOW);

      // If it is too long (ball not detected after 10 seconds), we counts 1 points not to block the game
      CptBallFall = millis() - TimeRefBallFall;
      if (CptBallFall > 10000)
        Switch[8] = LOW;
    
      for (int i = 0; i <= 15; i++){
        // Limit Switch pressed
        if (!Switch[i]){

          PrevScore = Scores[CurPlayer];

          // + or - ?
          if (ColorSelect == 1)
            Scores[CurPlayer] += Points[i];  // Add points
          else{
            Scores[CurPlayer] -= Points[i];  // Subtract points
            if (Scores[CurPlayer] < 0)
              Scores[CurPlayer] = 0;         // Never less than zero
          }
          NbBallFall++;

          // Show player X score


          // Animation of score only il it's the last ball
          if (NbBallFall >= NbBallsSelect){
            // ++++
            if ((Scores[CurPlayer] - PrevScore) > 0){
              StepDispl = (Scores[CurPlayer] - PrevScore) < 30 ? 1 : 10;
              for (int i = PrevScore; i <= Scores[CurPlayer]; i+=StepDispl){
                sprintf(Tampon, "%d", (int) i);            
                printText(Tampon);
              }
            }
            // ---
            else{
              StepDispl = (PrevScore - Scores[CurPlayer]) < 30 ? 1 : 10;
              for (int i = PrevScore; i > Scores[CurPlayer]; i-=StepDispl){
                sprintf(Tampon, "%d", (int) i);            
                printText(Tampon);
              }
            }
          }
          else{
            sprintf(Tampon, "%d", (int) Scores[CurPlayer]);            
            printText(Tampon);
          }
          
          
          // End of round, All the balls fell
          if (NbBallFall >= NbBallsSelect){
            
            rgbStrip(0, 0, 255);  // Blue
            delay(1000); // Pause for show score
            
            CurPlayer++;
            if (CurPlayer > NbPlayer){
              CurTurn++;
              if (CurTurn > NbTurn)
                STEP = 80;  // En of game 
              else{
                CurPlayer = 1;
                STEP = 30;  // Next round
              }
            }
            else{
              NbBallGiven = 0; 
              STEP = 40;  // Next player    
            }        
          }

          delay(200); // Avoid double detection of same ball
        }
      }


      // Show player X score
      sprintf(Tampon, "%d", (int) Scores[CurPlayer]);            
      printText(Tampon);
      break;


    case 80:
      // Find the winner
      for (int i=1; i<=NbPlayer; i++){
        if (Scores[i] > ValWinner){
          IdWinner = i; 
          ValWinner = Scores[i];   
        }
      }

      sprintf(Tampon, TxtArray[8], (int) IdWinner); 
      if (bRestart = scrollText(bRestart, Tampon))
        STEP = 90;
      break;


    case 90:
      // Show all scores
      sprintf(Tampon, TxtArray[5], (int) FinalScore, (int) Scores[FinalScore]);

      if (bRestart = scrollText(bRestart, Tampon)){
        FinalScore++;
        if (FinalScore > NbPlayer)
          FinalScore = 1;  
      }
      
      if (BtOK.read() == MD_KeySwitch::KS_PRESS)
        STEP = 100;  
      break;

          
    case 100:
      // Reset all for start a new game
      if (bRestart = scrollText(bRestart, TxtArray[0]))
        STEP = 0;  
      break;
  }

  rgbStrip.show(); 
}
